package cn.com.qmth.scancentral.service.impl;

import cn.com.qmth.scancentral.bean.User;
import cn.com.qmth.scancentral.bean.card.CardFile;
import cn.com.qmth.scancentral.client.MarkingcloudApiClient;
import cn.com.qmth.scancentral.client.MarkingcloudProperties;
import cn.com.qmth.scancentral.dao.AnswerCardDao;
import cn.com.qmth.scancentral.entity.*;
import cn.com.qmth.scancentral.enums.CardSource;
import cn.com.qmth.scancentral.enums.ExamMode;
import cn.com.qmth.scancentral.enums.SystemMode;
import cn.com.qmth.scancentral.model.AnswerCardInfo;
import cn.com.qmth.scancentral.service.*;
import cn.com.qmth.scancentral.vo.AnswerCardVo;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.UpdateWrapper;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.jeffreyning.mybatisplus.service.MppServiceImpl;
import com.qmth.boot.core.exception.ParameterException;
import com.qmth.boot.core.exception.StatusException;
import com.qmth.boot.core.fss.store.FileStore;
import com.qmth.boot.core.retrofit.exception.RetrofitResponseError;
import com.qmth.boot.core.retrofit.utils.SignatureInfo;
import com.qmth.boot.core.retrofit.utils.UploadFile;
import com.qmth.boot.core.security.exception.AuthorizationException;
import com.qmth.boot.tools.signature.SignatureType;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

@Service
public class AnswerCardServiceImpl extends MppServiceImpl<AnswerCardDao, AnswerCardEntity>
        implements AnswerCardService {

    protected static final Logger log = LoggerFactory.getLogger(AnswerCardServiceImpl.class);

    @Autowired
    private PaperStructureService paperStructureService;

    @Autowired
    private ExamService examService;

    @Autowired
    private FileService fileService;

    @Autowired
    private FileStore fileStore;

    @Autowired
    private SubjectService subjectService;

    @Autowired
    private MarkingcloudApiClient apiClient;

    @Autowired
    private ScannerCardService scannerCardService;

    @Autowired
    private SessionService sessionService;

    @Autowired
    private MarkingcloudProperties properties;

    @Autowired
    private PaperService paperService;

    @Autowired
    private AdapteFileService adapteFileService;

    @Autowired
    private AnswerCardSubjectService answerCardSubjectService;

    @Override
    public int getCountByExam(Long examId) {
        QueryWrapper<AnswerCardEntity> wrapper = new QueryWrapper<>();
        LambdaQueryWrapper<AnswerCardEntity> lw = wrapper.lambda();
        lw.eq(AnswerCardEntity::getExamId, examId);
        return this.count(wrapper);
    }

    @Transactional
    @Override
    public AnswerCardEntity save(User user, Long examId, Integer number, String subjectCode, String parameter,
            String remark, Integer paperCount, Boolean singlePage, String md5, Integer dpi, MultipartFile file) {
        ExamEntity exam = examService.getById(examId);
        if (ExamMode.K12.equals(exam.getMode()) && StringUtils.isBlank(subjectCode)) {
            throw new ParameterException("K12模式不允许创建通卡");
        }
        if (StringUtils.isNotBlank(subjectCode)) {
            SubjectEntity sub = subjectService.findByExamIdAndCode(examId, subjectCode);
            if (sub == null) {
                throw new ParameterException("未找到科目,请先同步或创建考生获取科目");
            }
        }
        AnswerCardEntity entity = null;
        if (number != null) {
            int studentCount = paperService.getCountByExamAndCardNumber(examId, number);
            if (studentCount != 0) {
                throw new ParameterException("卡格式已被使用，无法修改");
            }
            entity = this.findByExamAndNumber(examId, number);
            if (StringUtils.isNotBlank(subjectCode) && !answerCardSubjectService
                    .checkSubject(examId, number, subjectCode)) {
                throw new ParameterException("卡格式与该科目无关联，无法修改");
            }
        }

        CardFile cardFile;
        byte[] fileData;
        String sliceConfig;
        List<String> sliceName = null;
        try {
            fileData = file.getBytes();
            //解析卡格式文件
            cardFile = parseCardFile(fileData);
            if (singlePage && paperCount != cardFile.getPages().size()) {
                throw new ParameterException("卡格式数量不一致");
            }
            if (!singlePage && paperCount * 2 != cardFile.getPages().size()) {
                throw new ParameterException("卡格式数量不一致");
            }
            //提取裁切坐标
            sliceConfig = cardFile.getSliceConfig().toString();
            if (ExamMode.CET.equals(exam.getMode())) {
                sliceName = cardFile.getSliceName();
            }
        } catch (IOException e) {
            throw new ParameterException("文件解析失败", e);
        }
        if (SystemMode.MARKINGCLOUD.equals(SystemMode.current())) {
            //            log.info("markingcloud card save: account={}, token={}", user.getAccount(), user.getMarkingCloudToken());
            SignatureInfo signatureInfo = new SignatureInfo(SignatureType.TOKEN, user.getAccount(),
                    user.getMarkingCloudToken());
            AnswerCardInfo answerCardInfo = apiClient
                    .saveAnswerCard(signatureInfo, examId, number, CardSource.CLIENT, subjectCode, parameter, remark,
                            paperCount, singlePage, false, sliceConfig, md5, dpi,
                            UploadFile.build("file", "", fileData));
            number = answerCardInfo.getNumber();
            entity = this.findByExamAndNumber(examId, number);
        } else {
            if (number != null) {
                entity = this.findByExamAndNumber(examId, number);
            } else {
                number = findMaxCardNumberByExamId(examId) + 1;
            }
        }
        String filePath;
        try {
            filePath = fileService.uploadAnswerCard(file.getInputStream(), md5, examId, number);
        } catch (IOException e) {
            e.printStackTrace();
            throw new ParameterException("文件上传失败", e);
        }
        if (entity == null) {
            entity = new AnswerCardEntity();
            entity.setExamId(examId);
            entity.setNumber(number);
            if (StringUtils.isNotBlank(subjectCode)) {
                AnswerCardSubjectEntity as = new AnswerCardSubjectEntity();
                as.setExamId(examId);
                as.setCardNumber(number);
                as.setSubjectCode(subjectCode);
                answerCardSubjectService.saveOrUpdateByMultiId(as);
            }
        }
        entity.setPaperCount(paperCount);
        entity.setSinglePage(singlePage);
        entity.setParameter(parameter);
        entity.setRemark(remark);
        entity.setDpi(dpi);
        entity.setSliceConfig(sliceConfig);
        entity.setPath(filePath);
        entity.setMd5(md5);
        entity.setSource(CardSource.CLIENT);
        entity.setNeedAdapte(false);
        entity.setSliceName(sliceName);
        this.saveOrUpdateByMultiId(entity);
        //提取客观题结构
        paperStructureService.updateByAnswerCard(cardFile, examId, number);
        return entity;
    }

    private Integer findMaxCardNumberByExamId(Long examId) {
        int number = 0;
        QueryWrapper<AnswerCardEntity> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(AnswerCardEntity::getExamId, examId);
        List<AnswerCardEntity> list = this.list(queryWrapper);
        for (AnswerCardEntity card : list) {
            if (number < card.getNumber()) {
                number = card.getNumber();
            }
        }
        return number;
    }

    @Transactional
    @Override
    public void delete(User user, Long examId, Integer number) {
        if (examId == null) {
            throw new ParameterException("examId 不能为空");
        }
        if (number == null) {
            throw new ParameterException("number 不能为空");
        }
        int studentCount = paperService.getCountByExamAndCardNumber(examId, number);
        if (studentCount != 0) {
            throw new ParameterException("卡格式已被使用，无法删除");
        }
        AnswerCardEntity a = this.findByExamAndNumber(examId, number);
        if (CardSource.WEB.equals(a.getSource())) {
            throw new ParameterException("电子卡格式无法删除");
        }
        UpdateWrapper<AnswerCardEntity> updateWrapper = new UpdateWrapper<>();
        updateWrapper.lambda().eq(AnswerCardEntity::getExamId, examId);
        updateWrapper.lambda().eq(AnswerCardEntity::getNumber, number);
        int count = this.baseMapper.delete(updateWrapper);
        if (count == 0) {
            throw new ParameterException("未找到卡格式");
        }
        if (SystemMode.MARKINGCLOUD.equals(SystemMode.current())) {
            SignatureInfo signatureInfo = new SignatureInfo(SignatureType.TOKEN, user.getAccount(),
                    user.getMarkingCloudToken());
            apiClient.deleteAnswerCard(signatureInfo, examId, number);
        }
        answerCardSubjectService.deleteByExamAndNumber(examId, number);
        paperStructureService.deleteByAnswerCard(examId, number);
        scannerCardService.deleteByExamAndCardNumber(examId, number);
    }

    @Override
    public List<AnswerCardVo> listByExamId(Long examId) {
        QueryWrapper<AnswerCardEntity> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(AnswerCardEntity::getExamId, examId);
        List<AnswerCardEntity> list = this.list(queryWrapper);
        List<AnswerCardVo> result = new ArrayList<>();
        for (AnswerCardEntity answerCardEntity : list) {
            AnswerCardVo card = new AnswerCardVo(answerCardEntity);
            card.setAdapteFile(adapteFileService.listByExamIdAndCardNumber(examId, answerCardEntity.getNumber()));
            card.setSubjectList(answerCardSubjectService.findByExamAndNumber(examId, answerCardEntity.getNumber()));
            result.add(card);
        }
        return result;
    }

    @Override
    public List<AnswerCardVo> listByExamIdForScanner(Long examId, User scanner, String device) {
        List<AnswerCardVo> list = listByExamId(examId);
        if (CollectionUtils.isNotEmpty(list)) {
            List<ScannerCardEntity> sc = scannerCardService.getListByExamAndDevice(examId, scanner.getAccount());
            if (CollectionUtils.isEmpty(sc)) {
                for (AnswerCardVo vo : list) {
                    vo.setEnable(true);
                }
            } else {
                Set<Integer> set = new HashSet<>();
                for (ScannerCardEntity e : sc) {
                    set.add(e.getCardNumber());
                }
                for (AnswerCardVo vo : list) {
                    if (set.contains(vo.getNumber())) {
                        vo.setEnable(true);
                    } else {
                        vo.setEnable(false);
                    }
                    vo.setAdapteFile(
                            adapteFileService.listByExamIdAndCardNumberForScanner(examId, vo.getNumber(), device));
                    vo.setSubjectList(answerCardSubjectService.findByExamAndNumber(examId, vo.getNumber()));
                }
            }
        }
        return list;
    }

    @Transactional
    @Override
    public void sync(User user, Long examId) {
        if (SystemMode.STANDALONE.equals(SystemMode.current())) {
            return;
        }
        ExamEntity exam = examService.getById(examId);
        SignatureInfo signatureInfo = new SignatureInfo(SignatureType.TOKEN, user.getAccount(),
                user.getMarkingCloudToken());
        try {
            List<AnswerCardInfo> yun;
            yun = apiClient.getAnswerCardInfo(signatureInfo, examId, null);
            Set<Integer> numbers = new HashSet<>();
            for (AnswerCardInfo answerCardInfo : yun) {
                log.info("answer card file url: " + properties.getFileServer() + answerCardInfo.getPath());
                numbers.add(answerCardInfo.getNumber());
                QueryWrapper<AnswerCardEntity> queryWrapper = new QueryWrapper<>();
                queryWrapper.lambda().eq(AnswerCardEntity::getExamId, examId);
                queryWrapper.lambda().eq(AnswerCardEntity::getNumber, answerCardInfo.getNumber());
                AnswerCardEntity entity = this.getOne(queryWrapper);
                if (entity == null) {
                    fileService.downLoadAnswerCard(examId, answerCardInfo.getNumber(), answerCardInfo.getMd5(),
                            properties.getFileServer() + answerCardInfo.getPath());
                    entity = new AnswerCardEntity();
                    entity.setExamId(examId);
                    entity.setNumber(answerCardInfo.getNumber());
                    entity.setCode(answerCardInfo.getCode());
                    entity.setDpi(answerCardInfo.getDpi() == 0 ? null : answerCardInfo.getDpi());
                    entity.setPaperCount(answerCardInfo.getPaperCount());
                    entity.setSinglePage(answerCardInfo.getSinglePage());
                    entity.setParameter(answerCardInfo.getParameter());
                    entity.setRemark(answerCardInfo.getRemark());
                    entity.setPath(fileService.getAnswerCardUri(examId, answerCardInfo.getNumber()));
                    entity.setMd5(answerCardInfo.getMd5());
                    entity.setSliceConfig(answerCardInfo.getSliceConfig());
                    entity.setSource(answerCardInfo.getSource());
                    entity.setNeedAdapte(answerCardInfo.getSource() == CardSource.WEB);
                    CardFile cardFile = new ObjectMapper().readValue(fileStore.read(entity.getPath()), CardFile.class);
                    if (ExamMode.CET.equals(exam.getMode())) {
                        entity.setSliceName(cardFile.getSliceName());
                    }
                    this.saveOrUpdateByMultiId(entity);
                    paperStructureService.updateByAnswerCard(cardFile, examId, entity.getNumber());
                } else if (!answerCardInfo.getMd5().equals(entity.getMd5())) {
                    int studentCount = paperService.getCountByExamAndCardNumber(examId, answerCardInfo.getNumber());
                    if (studentCount == 0) {
                        if (entity.getSource().equals(CardSource.WEB)) {
                            adapteFileService.deleteByExamIdAndCardNumber(examId, entity.getNumber());
                        }
                        fileService.downLoadAnswerCard(examId, answerCardInfo.getNumber(), answerCardInfo.getMd5(),
                                properties.getFileServer() + answerCardInfo.getPath());
                        CardFile cardFile = new ObjectMapper()
                                .readValue(fileStore.read(entity.getPath()), CardFile.class);
                        entity.setPaperCount(answerCardInfo.getPaperCount());
                        entity.setSinglePage(answerCardInfo.getSinglePage());
                        entity.setMd5(answerCardInfo.getMd5());
                        entity.setSliceConfig(answerCardInfo.getSliceConfig());
                        if (ExamMode.CET.equals(exam.getMode())) {
                            entity.setSliceName(cardFile.getSliceName());
                        }
                        entity.setRemark(answerCardInfo.getRemark());
                        entity.setParameter(answerCardInfo.getParameter());
                        this.saveOrUpdateByMultiId(entity);
                        paperStructureService.updateByAnswerCard(cardFile, examId, entity.getNumber());
                    }
                }
                //重置卡格式与科目绑定关系
                answerCardSubjectService.deleteByExamAndNumber(examId, answerCardInfo.getNumber());
                for (String subjectCode : answerCardInfo.getSubjectList()) {
                    AnswerCardSubjectEntity as = new AnswerCardSubjectEntity();
                    as.setExamId(examId);
                    as.setCardNumber(answerCardInfo.getNumber());
                    as.setSubjectCode(subjectCode);
                    answerCardSubjectService.saveOrUpdateByMultiId(as);
                }
            }
            QueryWrapper<AnswerCardEntity> queryWrapper = new QueryWrapper<>();
            queryWrapper.lambda().eq(AnswerCardEntity::getExamId, examId);
            List<AnswerCardEntity> list = this.list(queryWrapper);
            for (AnswerCardEntity answerCardEntity : list) {
                if (!numbers.contains(answerCardEntity.getNumber())) {
                    this.delete(user, examId, answerCardEntity.getNumber());
                }
            }
            examService.updateCardSyncTime(examId);
        } catch (RetrofitResponseError e) {
            if (e.getCode() == 401) {
                sessionService.userLogout(user);
                throw AuthorizationException.SIGNATURE_INVALID;
            } else {
                throw e;
            }
        } catch (Exception e) {
            throw new StatusException("answer card sync error: " + e.getMessage(), e);
        }
    }

    @Override
    public AnswerCardEntity findByExamAndNumber(Long examId, Integer number) {
        if (examId == null) {
            throw new ParameterException("examId不能为空");
        }
        if (number == null) {
            throw new ParameterException("number不能为空");
        }
        QueryWrapper<AnswerCardEntity> wrapper = new QueryWrapper<>();
        LambdaQueryWrapper<AnswerCardEntity> lw = wrapper.lambda();
        lw.eq(AnswerCardEntity::getExamId, examId);
        lw.eq(AnswerCardEntity::getNumber, number);
        return baseMapper.selectOne(wrapper);
    }

    /**
     * 解析卡格式文件中的裁切图坐标，用于云阅卷同步
     *
     * @param data
     */
    private CardFile parseCardFile(byte[] data) throws IOException {
        ObjectMapper objectMapper = new ObjectMapper();
        return objectMapper.readValue(data, CardFile.class);
    }

}
